﻿/****************************************************************************
**
** Copyright (C) 2013 Jolla Ltd.
** Contact: Petri M. Gerdt <petri.gerdt@jollamobile.com>
**
****************************************************************************/

import QtQuick 2.1
import org.nemomobile.lipstick 0.1
import org.nemomobile.notifications 1.0 as Nemo
import org.nemomobile.configuration 1.0
import Sailfish.Silica 1.0
import Sailfish.Silica.theme 1.0
import com.jolla.lipstick 0.1
import Sailfish.Bluetooth 1.0
import Sailfish.Ambience 1.0
import "ambiencepicker"
import "bluetooth"
import "main"
import "launcher"
import "lockscreen"
import "statusarea"
import "switcher"
import "scripts/desktop.js" as Desktop

ApplicationWindow {
    id:window
    cover: undefined
    _backgroundVisible: false

    function desktopInstance() {
        return Desktop.instance
    }

    ConfigurationValue {
        id: shortcutsConfig

        key: "/desktop/lipstick-jolla-home/lock_screen_shortcuts"
        defaultValue: []
    }

    initialPage: Component { AmbiencePage {
        id: desktop

        property bool closeApplicationEnabled: false
        property bool removeApplicationEnabled: false
        property bool ambienceSwitcherAllowed: dashboard.currentIndex == 1
                                               && !dashboard.moving
                                               && !pullDownMenu.active
                                               && !deviceIsLocked
                                               && !ambienceInfoActive
                                               && !closeApplicationEnabled
                                               && !launcher.openedChildFolder
        property alias housekeepingMenuActive: housekeeping.housekeepingMenuActive
        property alias animating: yBehavior.running
        property alias animationDuration: yBehavior.duration
        property alias switcher: switcher

        property bool peeking: !Lipstick.compositor.homeActive &&
                               !lipstickSettings.lockscreenVisible
        property bool isInPeekingPosition: dashboard.contentY == dashboard.peekingPosition && !yBehavior.running
        property bool isInSwitcherPosition: dashboard.contentY == dashboard.lockscreenAndStatusAreaHeight
        property bool inLauncher: dashboard.currentIndex > 1
        property bool screenIsLocked: lipstickSettings.lockscreenVisible
        property bool deviceIsLocked: deviceLock.state == 1
        property bool pinQueryCompleted

        property bool ambienceOnLockscreen
        property alias dashboardInstance: dashboard

        property real wallpaperRange: Math.max(desktop.height*2, (dashboard.contentHeight-desktop.height))

        Connections {
            target: Lipstick.compositor
            onHomeActiveChanged: if (Lipstick.compositor.homeActive && !lipstickSettings.lockscreenVisible) dashboard.snapToPage(1)
            onDisplayOff: {
                pullDownMenu.close(true)
                deviceLockUI.reset()
                if (!lipstickSettings.lockscreenVisible && deviceIsLocked) {
                    dashboard.snapToPage(0, true)
                }
            }
            onDisplayOn: {
                if (pinQueryAgent.isWindowVisible()) {
                    lipstickSettings.lockscreenVisible = false
                }
            }
            onEventsPeekingChanged: if (Lipstick.compositor.eventsPeeking) desktop.closeApplicationEnabled = false
        }

        onCloseApplicationEnabledChanged: {
            if (closeApplicationEnabled) {
                housekeeping.show()
            } else {
                housekeeping.close()
            }
            onEventsPeekingChanged: if (Lipstick.compositor.eventsPeeking) desktop.closeApplicationEnabled = false
        }

        onScreenIsLockedChanged: {
            if (screenIsLocked) {
                desktop.closeApplicationEnabled = false
            }
        }

        onIsInSwitcherPositionChanged: if (!isInSwitcherPosition) desktop.closeApplicationEnabled = false

        function fastPeek() {
            var dist = Math.abs(dashboard.contentY - dashboard.peekingPosition)
            if (dist > 0) {
                if (dist > height + screen.height) {
                    fastPeekAnimation.run()
                } else {
                    dashboard.snapToPage(-1)
                }
            }
        }

        function setForceTopWindowProcessId(pid) {
            lockscreen.forceTopWindowProcessId = pid
        }

        function _showInitialPinQuery() {
            // Don't show until device lock is done, to ensure ofono has started to determine whether PIN is required.
            if (pinQueryCompleted || deviceIsLocked) {
                return
            }
            pinQueryAgent.enabled = true

            // Don't show if firstLipstickRun=true as start-up wizard shows PIN query in that case.
            if (!firstLipstickRun && pinQueryAgent.isSimPinRequired()) {
                pinQueryAgent.requestSimPin()
                pinQueryCompleted = true
            }
        }

        WallpaperPath {
            id: wallpaperPath
            contentY: dashboard.contentY
            yMarginAtTop: pullDownMenu._activeHeight + pullDownMenu.spacing
            yAtPeekingPosition: dashboard.peekingPosition
            yAtSwitcher: dashboard.lockscreenAndStatusAreaHeight
            yAtBottom: wallpaperRange
            screenHeight: screen.height
            maxVerticalOffset: Lipstick.compositor.wallpaperSize.height - screen.height

            onVerticalOffsetChanged: Lipstick.compositor.backgroundVerticalOffset = verticalOffset
        }

        property real dimmedBackgroundValue: desktop.closeApplicationEnabled || desktop.removeApplicationEnabled ? 0.8 : 0.0
        Behavior on dimmedBackgroundValue { NumberAnimation { duration: 200 } }
        Binding {
            target: Lipstick.compositor
            property: "backgroundOpacity"
            value: lipstickSettings.lowPowerMode ? 0 : (1-Math.max(dimmedBackgroundValue, wallpaperPath.opacity))
        }

        SilicaFlickable {
            id: housekeeping

            property bool closePending: false
            property Item oldParent
            property bool housekeepingMenuActive: housekeepingMenu.active

            function show() {
                visible = true
                oldParent = switcher.parent
                switcher.parent = housekeeping.contentItem
            }

            function close() {
                if (visible) {
                    switcher.parent = oldParent
                    visible = false
                }

                if (housekeepingMenuActive) {
                    housekeepingMenu.hide()
                }
            }

            anchors.fill: parent
            contentHeight: switcher.height
            visible: false

            onHousekeepingMenuActiveChanged: {
                if (!housekeepingMenuActive && closePending) {
                    housekeeping.closePending = false
                    switcher.parent = housekeeping.oldParent
                    housekeeping.visible = false
                    switcher.closeAll()
                }
            }

            PullDownMenu {
                id: housekeepingMenu

                menuIndicator: MenuIndicator {}
                MenuItem {
                    //: Menu item in pull down menu, closes all open applications
                    //% "Close all"
                    text: qsTrId("lipstick-jolla-home-me-close_all")
                    onClicked: housekeeping.closePending = true
                }
            }
        }

        SilicaListView {
            id: dashboard

            property real lockscreenAndStatusAreaHeight: lockAndStatusDelegate.height
            property real peekingPosition: lockscreen.height - lockscreen.clockHeight
            property real fixPos: -1
            property bool coversVisible: Lipstick.active && currentIndex == 1

            function snapToPage(page, immediate) {
                if (page == 0 && pullDownMenu.active) {
                    return
                }

                var targetPosition

                switch (page) {
                case -1:
                    targetPosition = peekingPosition
                    break
                case 0:
                    targetPosition = 0
                    break
                default:
                    targetPosition = lockscreenAndStatusAreaHeight + (page - 1) * dashboard.height
                    break
                }

                animateTo(targetPosition, immediate ? 0 : (Math.abs(contentY - targetPosition) > desktop.height ? 250 : 200))
                if (immediate) {
                    currentIndex = page;

                    if (lipstickSettings.lockscreenVisible && currentIndex > 0) {
                        lipstickSettings.lockscreenVisible = false
                    } else if (!lipstickSettings.lockscreenVisible && currentIndex == 0) {
                        lipstickSettings.lockscreenVisible = true
                    }
                }
            }

            function animateTo(position, duration) {
                if (contentY != position) {
                    // If a flick isn't started, the fixup will start and won't be cancelled.
                    // Force the contentY to our start position in case the fixup animation starts.
                    fixPos = contentY

                    if (duration == 0) {
                        yBehavior.stop()
                        fixPos = position
                        contentY = position
                    } else {
                        yBehavior.from = contentY
                        yBehavior.to = position
                        yBehavior.duration = duration === undefined ? 200 : duration
                        yBehavior.restart()
                    }
                } else {
                    yBehavior.stop()
                }
            }

            width: window.width
            height: window.height
            snapMode: ListView.SnapOneItem
            highlightRangeMode: ListView.StrictlyEnforceRange
            cacheBuffer: height*model.count
            maximumFlickVelocity: 4000*Theme.pixelRatio
            highlightMoveDuration: 300
            pressDelay: 0
            interactive: !closeApplicationEnabled && !removeApplicationEnabled && !launcher.openedChildFolder && !desktop.animating
            quickScroll: false

            onMovingChanged: {
                if (lipstickSettings.lockscreenVisible && currentIndex > 0) {
                    lipstickSettings.lockscreenVisible = false
                } else if (!lipstickSettings.lockscreenVisible && contentY == 0) {
                    lipstickSettings.lockscreenVisible = true
                }
            }

            onContentYChanged: {
                if (lockscreen.pressed) {
                    // As long as the lockscreen is pressed, keep its Y position
                    contentY = lockscreen.startContentY
                } else if (deviceLock.state == 1 && contentY > dashboard.lockscreenAndStatusAreaHeight) {
                    contentY = lockscreenAndStatusAreaHeight
                } else {
                    if (fixPos >= 0) {
                        contentY = fixPos
                        fixPos = -1;
                    }
                }
            }

            onCoversVisibleChanged: {
                if (coversVisible) {
                    coverControl.status = Cover.Activating
                    coverControl.status = Cover.Active
                } else {
                    coverControl.status = Cover.Deactivating
                    coverControl.status = Cover.Inactive
                }
            }

            PullDownMenu {
                id: pullDownMenu
                menuIndicator: MenuIndicator {}

                // Ensure the menu is above the lock screen
                z: lockscreen.z + 2

                // Hide the menu when the notification text area is visible
                opacity: (lockscreen.notificationAreaIsOpen || lipstickSettings.lowPowerMode) ? 0 : 1
                Behavior on opacity { FadeAnimation {} }

                Repeater {
                    model: LauncherWatcherModel {
                        filePaths: shortcutsConfig.value
                        onFilePathsChanged: shortcutsConfig.value = filePaths
                    }

                    MenuItem {
                        visible: model.object.isValid
                        text: model.object.title
                        onClicked: {
                            pullDownMenu.cancelBounceBack()
                            lipstickSettings.lockscreenVisible = false
                            if (!deviceIsLocked) {
                                desktop.fastPeek()
                            }
                            switcher.launchingApp(model.object)
                            model.object.launchApplication()
                        }
                    }
                }
                MenuItem {
                    text: profileControl.profile == "silent"
                            //: Menu item in pull down menu in lock screen, selecting item sets sounds on
                            //% "Switch sounds on"
                          ? qsTrId("lipstick-jolla-home-me-sounds_on")
                            //: Menu item in pull down menu in lock screen, selecting item sets sounds off
                            //% "Silence sounds"
                          : qsTrId("lipstick-jolla-home-me-sounds_off")
                    onClicked: { profileControl.profile = profileControl.profile == "silent" ? "ambience" : "silent" }
                }

                onActiveChanged: if (active) lipstickSettings.lockscreenVisible = true
            }

            Connections {
                target: desktop
                onPeekingChanged: {
                    if (!dashboard.dragging && !lipstickSettings.lockscreenVisible) {
                        if(!desktop.peeking) {
                            dashboard.animateTo(dashboard.lockscreenAndStatusAreaHeight)
                        } else {
                            // animate to peeking position
                            dashboard.animateTo(dashboard.peekingPosition)
                        }
                    }
                }
            }

            onDraggingChanged: {
                fixPos = -1;
                if (!dragging) {
                    // Must drag a larger portion of the screen to move to/from lock
                    if (contentY < dashboard.height*0.15 && contentY > 0) {
                        animateTo(0)
                    } else if (contentY > dashboard.height*0.85 && contentY < lockscreenAndStatusAreaHeight) {
                        animateTo(dashboard.height + statusArea.height)
                    }
                } else {
                    yBehavior.stop()
                }
            }

            SmoothedAnimation on contentY { id: yBehavior; duration: 200 }

            SequentialAnimation {
                id: fastPeekAnimation
                property real initialPos
                function run() {
                    initialPos = dashboard.contentY
                    start()
                }
                ParallelAnimation {
                    NumberAnimation { target: dashboard; property: "opacity"; to: 0.0; easing.type: Easing.InCubic; duration: 80 }
                    NumberAnimation {
                        target: dashboard; property: "contentY"; easing.type: Easing.InQuad; duration: 80
                        to: fastPeekAnimation.initialPos - launcher.cellHeight*1/3
                    }
                }
                ParallelAnimation {
                    NumberAnimation { target: dashboard; property: "opacity"; to: 1.0; easing.type: Easing.OutCubic; duration: 170 }
                    NumberAnimation {
                        target: dashboard; property: "contentY"; easing.type: Easing.OutQuad; duration: 170
                        from: dashboard.peekingPosition + launcher.cellHeight*2/3
                        to: dashboard.peekingPosition
                    }
                }
            }

            model: ListModel {}
            delegate: Item {
                id: delegateRoot
                width: comp.width
                height: comp.height
                Item {
                    id: container
                    parent: dashboard.contentItem
                    y: delegateRoot.y
                }
                Component.onCompleted: comp.parent = container
            }
        }

        Component.onCompleted: {
            Desktop.instance = desktop

            dashboard.model.append({ comp: lockAndStatusDelegate })
            dashboard.model.append({ comp: switcherAnchor })
            launcher.manageDummyPages()

            if (!lipstickSettings.lockscreenVisible) {
                dashboard.snapToPage(1, true)
            }
            _showInitialPinQuery()
        }

        onDeviceIsLockedChanged: {
            _showInitialPinQuery()
        }

        Column {
            id: lockAndStatusDelegate
            width: dashboard.width
            height: lockscreen.height + statusArea.height
            Lockscreen {
                id: lockscreen

                property real startX
                property real startContentY

                height: dashboard.height
                width: dashboard.width

                onPressed: {
                    startX = mouseX
                    startContentY = dashboard.contentY
                }

                onPositionChanged: {
                    if (Math.abs(mouseX - startX) > 3*Theme.startDragDistance) {
                        preventStealing = true
                        if (!scrollHintAnimation.running) {
                            // While the screen is pressed, the animation will be
                            // in the paused state, but that's okay, as we do not
                            // want the content to "detach" from the finger while
                            // it's touched. Once released, the animation will play.
                            scrollHintAnimation.start()
                        }
                    }
                }

                onClicked: {
                    // Only start if it's not already started (even if paused)
                    if (!scrollHintAnimation.running && !scrollHintAnimation.paused) {
                        scrollHintAnimation.start()
                    }
                }

                onCanceled: {
                    // User started to drag - stop the (paused) animation, as
                    // resuming is not useful once dragging has started
                    scrollHintAnimation.stop()
                    preventStealing = false
                }

                onReleased: preventStealing = false

                SequentialAnimation {
                    id: scrollHintAnimation

                    // Pause the animation while the user touches the screen
                    paused: running && lockscreen.pressed

                    NumberAnimation {
                        target: dashboard
                        property: "contentY"
                        to: dashboard.originY + Theme.itemSizeExtraSmall
                        duration: 400
                        easing.type: Easing.InOutQuad
                    }
                    NumberAnimation {
                        target: dashboard
                        property: "contentY"
                        to: dashboard.originY
                        duration: 400
                        easing.type: Easing.InOutQuad
                    }
                }
            }
            StatusArea {
                id: statusArea
                width: dashboard.width
            }
            visible: dashboard.contentY < lockAndStatusDelegate.height
        }

        Item {
            id: switcherAnchor
            width: dashboard.width
            height: dashboard.height
            opacity: dashboard.contentY > 0 && dashboard.contentY < lockAndStatusDelegate.height + switcherAnchor.height ? 1.0 : 0.0

            DeviceLock {
                id: deviceLockUI
                width: dashboard.width
                height: dashboard.height
                opacity: deviceLock.state == 1 ? 1 : 0
                enabled: deviceLock.state == 1
                areaVisible: dashboard.contentY >= lockAndStatusDelegate.height && enabled

                Behavior on opacity { FadeAnimation {} }

                Binding {
                   target: Lipstick.compositor
                   property: "bottomPeekAllowed"
                   value: deviceLockUI.peekAllowed
                }
            }

            Switcher {
                id: switcher
                width: dashboard.width
                opacity: deviceLock.state == 1 ? 0 : 1
                enabled: deviceLock.state != 1 && !desktop.animating

                Behavior on opacity { FadeAnimation {} }
            }
        }

        Item {
            id: launcherAnchor
            width: dashboard.width
            height: dashboard.height
            parent: switcherAnchor.parent
            y: switcherAnchor.y + switcherAnchor.height - (desktop.closeApplicationEnabled ? 0 : launcher.cellHeight)
            opacity: deviceLock.state > 0 ? 0 : 1
            enabled: opacity == 1

            Behavior on opacity { FadeAnimation {} }
            Behavior on y {
                SpringAnimation {
                    spring: 3
                    damping: 0.2
                    mass: 0.3
                }
            }

            MouseArea {
                width: parent.width
                height: (dashboard.model.count - 2) * dashboard.height + launcher.cellHeight

                Launcher {
                    id: launcher
                    property QtObject launcherModifierObject

                    onHeightChanged: manageDummyPages()

                    function manageDummyPages() {
                        if (dashboard.model.count >= 2 && dashboard.height > 0) {
                            // Create dummy pages to allow paging
                            var pageCount = Math.max(desktop.removeApplicationEnabled ? 1 : 0,
                                                     Math.ceil((height - cellHeight)/dashboard.height))
                            while (dashboard.model.count-2 < pageCount) {
                                var dummy = dummyDelegate.createObject(dashboard.contentItem)
                                dashboard.model.append({ comp: dummy })
                            }
                            while (dashboard.model.count-2 > pageCount) {
                                var obj = dashboard.model.get(dashboard.model.count-1)
                                dashboard.model.remove(dashboard.model.count-1)
                            }
                        }
                    }

                    onLauncherEditModeChanged: {
                        launcher.manageDummyPages()
                        if (launcherEditMode) {
                            if (dashboard.contentY < dashboard.lockscreenAndStatusAreaHeight + switcher.height) {
                                dashboard.animateTo(dashboard.lockscreenAndStatusAreaHeight
                                                    + dashboard.height - launcher.cellHeight)
                                showLauncherModifierTimer.start()
                            } else if (dashboard.contentY > dashboard.lockscreenAndStatusAreaHeight + switcher.height + launcher.height - launcher.cellHeight - desktop.height) {
                                dashboard.animateTo(dashboard.lockscreenAndStatusAreaHeight + switcher.height + Math.max(launcher.height - desktop.height, 0))
                                showLauncherModifierTimer.start()
                            } else {
                                showLauncherModifier()
                            }
                        } else {
                            if (showLauncherModifierTimer.running) {
                                showLauncherModifierTimer.stop()
                                return
                            }

                            var contentY = launcherModifierObject.contentY
                            launcherModifierObject.close()
                            launcherModifierObject.destroy()
                            launcherModifierObject = null
                            if (!desktop.screenIsLocked) {
                                var page = Math.min(2 + Math.round((contentY - launcher.cellHeight) / dashboard.height), dashboard.count-1)
                                dashboard.contentY = dashboard.lockscreenAndStatusAreaHeight
                                        + switcher.height
                                        + contentY
                                dashboard.visible = true
                                dashboard.snapToPage(page);
                            } else {
                                dashboard.visible = true
                            }
                        }
                    }

                    function showLauncherModifier() {
                        var yOffset = 0
                        if (dashboard.contentY >= dashboard.lockscreenAndStatusAreaHeight + switcher.height) {
                            yOffset = dashboard.contentY
                                - dashboard.lockscreenAndStatusAreaHeight - switcher.height
                        }
                        launcherModifierObject = launcherModifierComponent.createObject(desktop, {"contentY": yOffset})
                        dashboard.visible = false
                        launcherModifierObject.show()
                    }

                    Timer {
                        id: showLauncherModifierTimer

                        interval: yBehavior.duration
                        onTriggered: {
                            launcher.showLauncherModifier()
                        }
                    }
                }

                onPressAndHold: launcher.setEditMode(true)
            }
        }

        BluetoothAgent {
            id: bluetoothAgent
            name: "/com/jolla/bluetooth_system_agent"

            onDisplayPasskey: bluetoothPairing.displayPasskey(deviceAddress, deviceClass, deviceName, normalizePasskey(key))
            onRequestConfirmation: bluetoothPairing.requestPasskeyComparison(deviceAddress, deviceClass, deviceName, normalizePasskey(key))
            onRequestPasskey: bluetoothPairing.requestPasskeyEntry(deviceAddress, deviceClass, deviceName)
            onRequestPidCode: bluetoothPairing.requestPinEntry(deviceAddress, deviceClass, deviceName)
            onRequestAuthorization: bluetoothPairing.requestAuthorization(deviceAddress, deviceClass, deviceName, uuid)
            onCanceledRequest: bluetoothPairing.canceledRequest()

            Component.onCompleted: registerAsSystemAgent()
        }
        BluetoothPairingView {
            id: bluetoothPairing

            onPasskeyComparisonConfirmed: {
                bluetoothAgent.trustDeviceBeingPaired(trustDevice)
                bluetoothAgent.replyRequestConfirmation(true)
            }
            onPasskeyDisplayConfirmed: {
                bluetoothAgent.trustDeviceBeingPaired(trustDevice)
                bluetoothAgent.replyRequestConfirmation(true)
            }
            onPasskeyEntered: {
                bluetoothAgent.trustDeviceBeingPaired(trustDevice)
                bluetoothAgent.replyPasskey(passkey)
            }
            onPinEntered: {
                bluetoothAgent.trustDeviceBeingPaired(trustDevice)
                bluetoothAgent.replyRequestPidCode(pidCode)
            }
            onAuthorizationResponse: {
                bluetoothAgent.trustDeviceBeingPaired(trustDevice)
                bluetoothAgent.replyRequestAuthorization(authorize)
            }
            onPairingCanceled: {
                bluetoothAgent.replyRequestConfirmation(false)
            }
        }

        BluetoothObexAgent { }
        BluetoothSession {
            onTurningBluetoothOn: bluetoothEnabledNotification.publish()
        }

        Nemo.Notification {
            id: bluetoothEnabledNotification
            category: "x-jolla.lipstick.bluetoothenabled"
            //% "Turning Bluetooth on"
            previewBody: qsTrId("lipstick-jolla-home-la-bluetoothon")
        }

        PinQueryAgent {
            id: pinQueryAgent
        }

        Component {
            id: dummyDelegate
            Item {
                width: dashboard.width
                height: dashboard.height
            }
        }

        Component {
            id: launcherModifierComponent

            Item {
                property bool closePending: false
                property Item oldParent
                property alias contentY: launcherModifier.contentY
                property alias contentHeight: launcherModifier.contentHeight

                function show() {
                    if (!oldParent) {
                        oldParent = launcher.parent
                        launcher.parent = launcherModifier.contentItem
                    }
                    visible = true
                }

                function close() {
                    if (oldParent) {
                        launcher.parent = oldParent
                        oldParent = null
                    }
                    visible = false
                }

                function scroll(up) {
                    contentYAnimation.to = up ? 0 : launcherModifier.contentHeight - desktop.height
                    contentYAnimation.duration = Math.abs(launcherModifier.contentY - contentYAnimation.to)
                    contentYAnimation.start()
                }

                function stopScrolling() {
                    contentYAnimation.stop()
                }

                anchors.fill: parent
                visible: false

                NumberAnimation {
                    id: contentYAnimation

                    target: launcherModifier
                    property: "contentY"
                }

                Column {
                    y: -launcherModifier.contentY
                    visible: parent.visible

                    Repeater {
                        model: (Math.ceil(launcherModifier.contentHeight / desktop.height) + 1)
                        Rectangle {
                            height: index === 0 ? launcher.cellHeight : desktop.height
                            width: desktop.width
                            opacity: index % 2 === 1 ? 0 : 0.3
                            color: Theme.highlightBackgroundColor
                        }
                    }
                }

                SilicaFlickable {
                    id: launcherModifier

                    anchors.fill: parent
                    contentHeight: Math.max(launcher.height, desktop.height + launcher.cellHeight)
                    visible: parent.visible

                    MouseArea {
                        height: Math.max(launcher.height, desktop.height + launcher.cellHeight)
                        width: parent.width
                        enabled: launcher.launcherEditMode
                        onClicked: launcher.setEditMode(false)
                    }
                }
            }
        }
    } }
}
